#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include "matrix.h"

m_double* new_matrix(int li,int co) {
   m_double* mat = NULL;
   if (li > 0 && co > 0) {
      mat =(m_double*) malloc(sizeof(m_double));
      mat->li = li; mat->co = co;
      double* data = (double*) malloc(li*co*sizeof(double));
      mat->content = (double**) malloc(li*sizeof(double*));
      for (int i=0;i<li;i++) mat->content[i] = data+i*co;
   }
   return mat;
}

void resize_line(m_double* mat,int li) {
   if (NULL != mat && li > 0) {
      mat->content = realloc(mat->content,li*sizeof(double*));
      mat->content[0] = realloc(mat->content[0],li*(mat->co)*sizeof(double));
      for (int i=0;i<li;i++) { 
         mat->content[i] = mat->content[0]+i*(mat->co);
         for (int j=0;j<mat->co;j++) {
            if (i >= mat->li) mat->content[i][j] = 0.0;
         }
      }               
      mat->li = li;
   }
}

void resize_column(m_double* mat,int co) {
   if (NULL != mat && co > 0) {
      int sign = (mat->co-co)/abs(mat->co-co);
      int first = (1+sign)/2 + ((mat->li)-1)*(1-sign)/2;
      int last  = (1-sign)/2 + ((mat->li)-1)*(1+sign)/2;
      
      if (co > mat->co) { //sign == -1
         mat->content[0] = realloc(mat->content[0],(mat->li)*co*sizeof(double));
         for (int i=1;i<mat->li;i++) mat->content[i] = mat->content[0]+i*co;
      }
      for (int i=first;sign*i<=sign*last;i+=sign) {
         memmove(
            mat->content[0]+i*co,
            mat->content[0]+i*(mat->co),
            (mat->co)*sizeof(double)
         ); 
      }        
      if (co < mat->co) { //sign == +1
         mat->content[0] = realloc(mat->content[0],(mat->li)*co*sizeof(double));
         for (int i=1;i<mat->li;i++) mat->content[i] = mat->content[0]+i*co;
      }
      
      for (int i=0;i<mat->li;i++)  
         for (int j=mat->co;j<co;j++) 
            mat->content[i][j] = 0.0; 
      
      mat->co = co;
   }
}

void resize(m_double* mat,int li,int co) {
   resize_column(mat,co);
   resize_line(mat,li);
}
   
m_double* transpose(m_double* m) {
   m_double* transp = NULL;
   if (NULL != m) {
      transp = new_matrix(m->co,m->li);
      for (int i=0;i<m->li;i++) 
         for (int j=0;j<m->co;j++) 
            transp->content[j][i] = m->content[i][j];
   }
   return transp;
}

m_double* rnd(int li,int co,
              double inf,double sup) {
   m_double* mat = new_matrix(li,co);
   if (NULL != mat) {
      for (int i=0;i<li;i++)
         for (int j=0;j<co;j++)
            mat->content[i][j] = inf+drand48()*(sup-inf);
   }
   return mat;
}

m_double* rnd_int(int li,int co,
                 int inf,int sup) {
   m_double* mat = new_matrix(li,co);
   if (NULL != mat) {
      for (int i=0;i<li;i++)
         for (int j=0;j<co;j++)
            mat->content[i][j] = inf+rand()%(sup-inf);
   }
   return mat;
}

m_double* add(m_double* m1,m_double* m2) {
   m_double* res = NULL;
   if (m1->li == m2->li && m1->co == m2->co) {
      res = new_matrix(m1->li,m1->co);
      for (int i=0;i<res->li;i++) 
         for (int j=0;j<res->co;j++) 
            res->content[i][j] = m1->content[i][j]+m2->content[i][j];
   }
   return res;
}

m_double* mult(m_double* m1,m_double* m2) {
   m_double* res = NULL;
   if (m1->co == m2->li) {
      res = new_matrix(m1->li,m2->co);
      for (int i=0;i<res->li;i++) 
         for (int j=0;j<res->co;j++) {
            res->content[i][j] = 0;
            for (int k=0;k<m1->co;k++)
               res->content[i][j] += m1->content[i][k]*m2->content[k][j];
         }
   }
   return res;
}

void free_matrix(m_double* m) {
   free(m->content[0]);
   free(m->content);
}

void format(m_double* m) {
   if (NULL != m) {
      printf("[");
      for (int i=0;i<m->li;i++) {
         for (int j=0;j<m->co;j++) printf("%f ",m->content[i][j]);
         if (i+1 == m->li) printf("]\n");
         else printf("; ");
      }
   } else printf("\n");
}

m_double* from_file(const char* filename) {
   FILE *fp = fopen(filename,"r+");
   if (NULL == fp) {
      fprintf(stderr,"%s, file not found\n",filename);
      exit(1);
   }
   int li = 0, co = 0, nb = 0;
   double* data = NULL;
   char* line = NULL;
   size_t n;
	while (getline(&line,&n,fp) != -1) {
	   char* ptr = line;
	   int ret, old = -1, nb_car;
	   float val;
	   while (1 == (ret = sscanf(ptr,"%f%n",&val,&nb_car))) {
         ptr += nb_car;   
         if (li == 0) co++;           
         data = (double*)realloc(data,(li+1)*co*sizeof(double));         
         data[nb++] = val;
         old = ret;
      }
      if (1 == ret || 1 == old) {
         old = -1;
         li++;
      }      
   }
   free(line);
   fclose(fp);
   
   m_double* mat = new_matrix(li,co);
   memcpy(mat->content[0],data,li*co*sizeof(double));
   free(data);
   return mat;
}

void print(m_double* m,char* filename) {
   FILE *fp = stdout;
   if (NULL != filename) fp = fopen(filename,"w+");
   if (NULL != m) { 
      for (int i=0;i<m->li;i++) {
         for (int j=0;j<m->co;j++) 
            fprintf(fp,"%d ",(int)(m->content[i][j]));
         fprintf(fp,"\n");
      }
   } else fprintf(fp,"\n");
}
